/* famplayer.c
 * GTK based Video Player for the iPAQ 
 */

/* Current design :
 * 
 * The decoder plugins produce video frames as packed 24-bit RGB buffer
 * We then just use  GdkRGB functions to put it on a GtkDrawingArea.
 * This seems fast enough (uses shared memory).
 *
 * Audio samples are decoded and given to us by the plugin.
 * The number of samples decoded per iteration is decided by the time 
 * required to decode the last frame and the time it takes to decode an audio
 * sample.
 * 
 * We do not enter the gtk_main() loop.
 * Instead we call gtk_main_iteration in our processing loop.
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>    
#include <string.h>
#include <sys/time.h>

#include <gtk/gtk.h>

#include "famaudio.h"
#include "famdecoderplugin.h"

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

/* PIXMAPS_DIR is picked up from compilation flags */
#define PLAY_XPM     PIXMAPS_DIR"/play.xpm"
#define PAUSE_XPM    PIXMAPS_DIR"/pause.xpm"

#define PLAY_STRING  "PLAY"
#define PAUSE_STRING "PAUSE"

#define MAX_DIM 320

/* Bytes perl pixel for the pixel buffer */
#define BYTES_PER_PIXEL      3

/* No. of samples for benchmarking audio decode rate */
#define TEST_NUM_SAMPLES     8192

/* Limits for the decode buffer size in samples */
#define MIN_AUDIO_SAMPLES    1024
#define MAX_AUDIO_SAMPLES    16384
#define STEP_AUDIO_SAMPLES   64

#define AUDIO_SAMPLE_SIZE    2048

/* For now keep dithering level fixed */
/* Doesn't seem to have too much effect on frame rate - So have it at MAX */
#define DITHER_LEVEL GDK_RGB_DITHER_MAX

/** Functional Macros **/
/* Decoder and draw one frame of video */
#define DRAW_VIDEO_FRAME  (decoder->readVideoFrame)(m_fp, \
                            (unsigned char **)pixel_rows, \
	                    0, 0, video_w, video_h, \
			    video_w, video_h); \
                          \
                     	  update_rect.x      = video_x;\
                          update_rect.y      = video_x;\
                      	  update_rect.width  = video_w;\
                     	  update_rect.height = video_h;\
	                  \
                 	  gtk_widget_draw(drawing_area, &update_rect);

/* Decode Audio */
#define DECODE_AUDIO(m_fp, buffer, num_samples) \
                          (decoder->readAudioSamples)(m_fp, \
                          buffer, num_samples) 
	  
/* Do one GTK Iteration */
#define GTK_ITERATION_DO  while (gtk_events_pending()) {\
	                    gtk_main_iteration();\
	                  }      

/* Common */
char    *file_path;
FamDecoderPlugin *plugin; /* Plugin Handle  */
FamDecoder *decoder;      /* Decoder object */
const char *decoder_name; /* Decoder name string */
void    *m_fp;            /* Opaque Media Decoder Handle */
int done = 0;
struct timeval start_tick;
long total_time, current_time;


enum { 
  PAUSED = 0,
  PLAYING
} player_status;

#define IS_PLAYING(status) (status == PLAYING)

/* GTK */
GtkWidget *drawing_area;   /* Video drawing Area     */
GdkRectangle update_rect;  /* Video update rectangle */
GtkWidget *play, *pause, *prev, *next, *playlist, *about;
GtkAdjustment *adjustment; /* For position indication */

/* Video */
int     have_video    = 0;
unsigned char *pixbuf = NULL;      /* Pixel Buffer */
unsigned char **pixel_rows = NULL; /* Pointer array - Rows of pixels */
int     origv_w, origv_h;          /* Original video dimensions */
int     video_w = 0, video_h = 0;  /* Video Dimensions that fits on screen */
int     video_x, video_y;
float   frate;            
long    fdecode;    
long    total_frames, current_frame, prev_frame;
long    ref_start;   /* Time when video started playing(maybe after pausing) */

/* Audio */
int   have_audio = 0;
int   audio_fd;
short *mixbuf;       /* Audio buffer */
long  test_decode;   /* Fixed time taken to decode TEST_NUM_SAMPLES */
long  adecode;       /* Audio decode time for a given audio buffer size */
long  agap;          /* Time it takes play the audio buffer */
AudioSpec spec;      /* Audio device parameters */  
long total_samples;  /* Length of the audio stream (number of samples) */

/* Usage */
void usage(char *argv[])
{
  printf("Usage : %s <Media Filename>\n", argv[0]);
}

/* Check whether a file can be opened for reading */
int fileExists(char *path)
{
  FILE *fp;
  
  if ((fp = fopen(path, "r")) == NULL) {
    return 0;
  }
  else {
    fclose(fp);
    return 1;
  }
}


/* Starting ticks value */
void startTicks()
{
  gettimeofday(&start_tick, 0);
}


/* Get the time in milliseconds from the start of the program */
long getTicks() 
{
  struct timeval now;

  gettimeofday(&now, 0);
  return ((now.tv_sec - start_tick.tv_sec)*1000 
	  + ((now.tv_usec - start_tick.tv_usec)/1000));  
}


/* Get the position in the stream
 * Returns frame number if video is present
 * Otherwise returns audio sample count
 */
int getPosition()
{
  long pos;

  if (have_video) {    
    (decoder->getFrame)(m_fp, &pos);
  }
  else {
    (decoder->getSample)(m_fp, &pos);
  }  

  return pos;
}

/* Seek to the given position 
 * If video is present the position is taken to be sample count
 * If only audio is present it is taken as sample count
 */
void setPosition(float value)
{
  long frame, sample;

  /* Can't seek - quietly return */
  if (!(decoder->isSeekable)(m_fp)) {
    return;
  }

  if ((have_video) && (have_audio)) { /* Both audio and video */

    frame = (long)(value);
    if (frame < 0)
      frame = 0;

#ifdef __DEBUG
    printf("Setting frame : %ld\n",frame);
#endif

    (decoder->setFrame)(m_fp, frame); 
    
    sample = (long)((double)(frame) / frate * spec.freq);

#ifdef __DEBUG
    printf("Setting sample : %ld\n", sample);
#endif

    (decoder->setSample)(m_fp, sample);
  }
  else if (have_video) {             /* Only video */

    frame = (long)(value);
    if (frame < 0)
      frame = 0;

#ifdef __DEBUG
    printf("Setting frame : %ld\n",frame);
#endif

    (decoder->setFrame)(m_fp, frame); 
  }
  else {                            /* Only audio */ 
    
    sample = (long)(value);

#ifdef __DEBUG
    printf("Setting sample : %ld\n", sample);
#endif

    (decoder->setSample)(m_fp, sample);
  }  
}


/** The following are GTK realted stuff **/

/* Main Window destroy event */
void quit() 
{
  done = 1;
}

/* Create a new hbox with an image packed into it and return the box
 */
GtkWidget *xpm_box( GtkWidget *parent,
		    gchar     *xpm_filename)
{
  GtkWidget *box1;
  GtkWidget *pixmapwid;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
  GtkStyle  *style;  
  
  /* Create box for xpm and label */
  box1 = gtk_hbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (box1), 2);
  
  /* Get the style of the button to get the
   * background color. */
  style = gtk_widget_get_style(parent);
  
  /* Now on to the xpm stuff */
  pixmap = gdk_pixmap_create_from_xpm (parent->window, &mask,
				       &style->bg[GTK_STATE_NORMAL],
				       xpm_filename);
  pixmapwid = gtk_pixmap_new (pixmap, mask);
    
  /* Pack the pixmap and label into the box */
  gtk_box_pack_start (GTK_BOX (box1), pixmapwid, FALSE, FALSE, 3);
    
  gtk_widget_show(pixmapwid);
  return(box1);
}


/* Redraw the screen from the RGB Array */
/* Currently draws the entire frame everytime */
/*** Review ***/
static gint expose_event( GtkWidget      *widget,
                          GdkEventExpose *event )
{
  /* Draw the current frame onto the widget */
  gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
		      video_x, video_y, video_w, video_h,
		      DITHER_LEVEL, pixbuf, 
		      video_w * BYTES_PER_PIXEL);
  
  return FALSE;
}

/* Create a new RGB pixel array of the appropriate size */
/* This is used to handle XRANDR changes during playback */
static gint configure_event( GtkWidget         *widget,
                             GdkEventConfigure *event )
{
  int i, old_w, old_h;

  old_w = video_w;
  old_h = video_h;  

  /* Fit the video within the drawing area */
  video_w = MIN(origv_w, widget->allocation.width);
  video_h = MIN(origv_h, widget->allocation.height);  

  /* Center the video image depending on the size of the drawing area */
  video_x = (widget->allocation.width  - video_w)/2;
  video_y = (widget->allocation.height - video_h)/2;    

  
  /* Reallocate RGB buffer if dimensions have changed */
  if ((old_w != video_w) || (old_h != video_h)) {

    /* Second time resize?? - Free last time's stuff */
    if (pixbuf) {
      free(pixbuf);
      free(pixel_rows);
    }
    
    /* Allocate pixel buffer */
    pixbuf = (unsigned char *)malloc(video_w * video_h * BYTES_PER_PIXEL);
    
    /* Present a pixel array as expected by the decoder */
    pixel_rows = (unsigned char **) calloc(video_h, sizeof(unsigned char *));
    pixel_rows[0] = (unsigned char *)((char *)pixbuf);   
    for (i = 1;i < video_h;i++) {
      pixel_rows[i] = pixel_rows[i-1] + (BYTES_PER_PIXEL * video_w);
    }    
  }
  
}


/* Synch Audio and Video streams
 */
void syncAV(void)
{
  long frame;
  long sample;
  
  if ((decoder->isSeekable)(m_fp)) {
    (decoder->getSample)(m_fp, &sample);
    frame = (long) ((sample * (double)frate)/(spec.freq));
    (decoder->setFrame)(m_fp, frame); 
  }

#ifdef __DEBUG
  printf("syncAV\n");
  printf("Sample : %ld\n", sample);
  printf("Frame  : %ld\n", frame);
#endif
}


void setPlayerStatus(int status) 
{
  if (status == PLAYING) {

    /* Set player status to PLAYING */
    player_status = PLAYING;

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(play), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pause), FALSE);    

    if (have_audio) {
      /* Re-open audio device */
      if ((audio_fd = openAudio(&spec)) < 0) {
	printf("Could not open audio device - Disabling audio.\n");
	have_audio = 0; /* Try to continue without sound */
      }       
    }

    /* Sync Audio and video */
    /* This is required beacuse of the large sample size we have to use 
     * for audio. Audio might have spilt above the video position. 
     */
    if (have_video  && have_audio) {
      syncAV();
    }

    if (have_video) {
      /* Reset video clock */
      prev_frame = -1;
    }    
  }
  else {
    player_status = PAUSED;
    
    closeAudio(audio_fd);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(play), FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pause), TRUE);       
  }
}


/* Handle Play and Pause button clicks */
void play_status(GtkWidget *widget,
		 gpointer   data)
{
  char *action = (char *)data;

  /* Do something only if there is a change in status */
  if (!strcmp(action, PLAY_STRING)) {

    if ((GTK_TOGGLE_BUTTON (widget)->active) && (player_status != PLAYING)) {
      setPlayerStatus(PLAYING);
    }
    else if (!(GTK_TOGGLE_BUTTON (widget)->active) && 
	     (player_status == PLAYING)) { /* Play pressed while playing */
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(play), TRUE);
    }

  }
  else if (!strcmp(action, PAUSE_STRING)) {
    if ((GTK_TOGGLE_BUTTON (widget)->active) && (player_status != PAUSED)) {
      setPlayerStatus(PAUSED);
    }
    else if (!(GTK_TOGGLE_BUTTON (widget)->active) && 
	     (player_status == PAUSED)) { /* Pause pressed while paused */
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pause), TRUE);
    }
  }
}


/* Handler for change in horizontal slider */
void slider_moved(GtkAdjustment *adj)
{ 
  /* Pause the video and audio */
  setPlayerStatus(PAUSED); 

  /* Value is frame count in case video is there 
   * Sample count in case of audio only streams 
   */
  if (have_video) {    
        
    /* Set one frame back and draw one frame */
    setPosition(adj->value - 1);      

    /*** Work around a problem with libmpeg3 
     *** Dead locks around here for video only streams 
     *** Ideally we should know nothing about the plugins
     ***/
    if ((!strcmp(decoder_name, "libmpeg3")) && (!have_audio)) {
      return;
    }

    /* Draw the current frame */
    DRAW_VIDEO_FRAME;
  }
  else { /* Only audio */
    setPosition(adj->value);      
  }
}


/* MAIN */
int main(int   argc, char *argv[])
{
  /* GTK Widgets */
  GtkWidget *window;
  GtkWidget *vbox;
  /* Drawing area is global */  
  GtkWidget *seperator;
  GtkWidget *slider;    
  GtkWidget *hbox;
  /* Player status buttons(play, pause etc.) are global */
  GtkWidget *box1;
  float upper, step, page, pos; /* For the slider */

  long  frames, fstart;
  
  /* Timing stuff */
  long start_time, stop_time;
  int   old_write_samples;
  
  /* Program starts here ! */
  startTicks();

  /* Let GTK take whetever it wants from the command line */
  gtk_init (&argc, &argv);
  
  /* Initialize GdkRGB */
  gdk_rgb_init();
  
  /* Get filename */
  if (argc != 2) {
    usage(argv);
    exit(2);
  }
  file_path = argv[1];
  
  /* Check whether the file exists */
  if (!fileExists(file_path)) {
    printf("%s : No such file or directory\n", file_path);
    exit(1);
  }
  
  /* Load the appropriate plugin to play the file */
  if ((plugin = loadDecoderPlugin(file_path)) == NULL) {  
    printf("%s : No supporting Decoder found\n", file_path);
    exit(1);
  }
  decoder = plugin->decoder; /* The decoder object */
  decoder_name = (decoder->decoderName)();

  printf("%s : \n", file_path);
  
  /* Open the file */
  m_fp = (decoder->openFile)(file_path);
  if (m_fp == NULL) {
    printf("%s : Could not open for some reason\n", file_path);
    exit(1);
  }
  
  /* Get video stream Info */
  if ((decoder->hasVideo)(m_fp)) {
    
    int i;   
    
    have_video = 1;
    
    /* Get the video width and height */
    (decoder->getVideoSize)(m_fp, &origv_w, &origv_h); 

    /* Get the frame rate - Used for video synchronization calculations */
    (decoder->getVideoFrameRate)(m_fp, &frate);
    
    /* And total no. of frames */
    (decoder->getVideoFrames)(m_fp, &total_frames);
    
    /* Caculate video time */
    total_time = (long)((float)(total_frames)/frate * 1000);    
    
    printf("  Video Size            = %dx%d\n", origv_w, origv_h);
    printf("  Frame Rate            = %3.2f\n", frate);
    printf("  Number of frames      = %ld\n",   total_frames);   
  }
  else {
    printf("No video streams found\n");
  }
  
  /* Get Audio Stream Info */
  if ((decoder->hasAudio)(m_fp)) {
    
    short *test_stream;

    have_audio = 1;
    
    /* Get the audio sample rate */
    (decoder->getAudioSampleRate)(m_fp, &spec.freq);

    /* Get the number of channels */
    (decoder->getAudioChannels)(m_fp, &spec.channels);
    if (spec.channels > 2)
      spec.channels = 2;
    
    /* Total no. of 16-bit samples in the stream */
    (decoder->getAudioSamples)(m_fp, &total_samples);

    /** Calculate time taken to decode TEST_NUM_SAMPLES samples of audio **/
    
    /* Allocate sound mixing buffer */
    mixbuf = (short *)malloc(MAX_AUDIO_SAMPLES * spec.channels * SAMPLE_SIZE);
    
    /* Start with some defaults and adjust as we go along */
    spec.samples = AUDIO_SAMPLE_SIZE;
    spec.write_samples = TEST_NUM_SAMPLES;
    spec.max_write_samples = MAX_AUDIO_SAMPLES;

    
    /* If video is also there calculate audio decode time 
     * for time sharing between and audio and video 
     */
    if (have_video) {

      start_time = getTicks();
      DECODE_AUDIO(m_fp, mixbuf, spec.write_samples);
      stop_time  = getTicks();
      
      /* Record the time taken to decode TEST_NUM_SAMPLES */
      test_decode = stop_time - start_time; /* msec */
      
      /* Rewind */
      if ((decoder->isSeekable)(m_fp)) {
	(decoder->setSample)(m_fp, 0);
      }
      else { /* Re-open if seeking is not available :-P */
	(decoder->closeHandle)(m_fp);
	m_fp = (decoder->openFile)(file_path);	
	/*** Do error checking again here ?? ***/
      }
    }


    /* Calculate total time */
    total_time = (long)((float)(total_samples)/spec.freq * 1000);    
    
    /* Print some info */
    printf("  No. of Audio channels = %d\n", spec.channels);
    printf("  Audio Sample Rate     = %3d Hz\n", spec.freq);    
    printf("  No. of samples        = %ld\n", total_samples);
    printf("  Length                = %3.2f sec\n", (float)(total_time)/1000);
  }
  else {
    printf("  Length                = %3.2f sec\n", (float)(total_time)/1000);
    printf("No audio channels found\n");
  }
  
  /* GTK Window stuff */
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_name (window, "Famplayer");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  
  /* Destroy handler */
  gtk_signal_connect (GTK_OBJECT (window), "destroy",
		      GTK_SIGNAL_FUNC (quit), NULL);
  

  /* Create the drawing area */
  drawing_area = gtk_drawing_area_new ();

  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), MAX_DIM, MAX_DIM);
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);
  
  if (have_video) {

    /* Signals used to handle the Drawing Area */        
    gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
			(GtkSignalFunc) expose_event, NULL);  
    
    gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
			(GtkSignalFunc) configure_event, NULL);
    
    gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK);    
    
  }  
  gtk_widget_show (drawing_area);
  
  /* Horizontal seperator */
  seperator = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), seperator, FALSE, FALSE, 3);
  gtk_widget_show (seperator);

  /* Create the slider bar */
  if (have_video) {
    step  = 1;  /* 1 frame   */
    page  = 10; /* 10 frames */ 
    upper = total_frames + page;	
  }
  else {     /* Only audio */
    step  = AUDIO_SAMPLE_SIZE;
    page  = step * 10; 
    upper = total_samples + page;	
  }

  adjustment = (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, upper, 
						     step, page, page);

  slider = gtk_hscale_new(GTK_ADJUSTMENT(adjustment));
  gtk_scale_set_draw_value(GTK_SCALE(slider), FALSE);
  GTK_WIDGET_UNSET_FLAGS (slider, GTK_CAN_FOCUS);
  gtk_range_set_update_policy(GTK_RANGE(slider), GTK_UPDATE_CONTINUOUS);
  
  /* Allow seeking thro' slider only when the decoder plugin allows seeking */
  if ((decoder->isSeekable)(m_fp)) {
    gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
			GTK_SIGNAL_FUNC (slider_moved), NULL);
  }
  
  gtk_box_pack_start (GTK_BOX (vbox), slider, FALSE, FALSE, 0);
  gtk_widget_show (slider);
  
  /* Player controls - Play, Pause, Prev, Next, Playlist and about buttons */
  /* A horizolntal row of buttons */
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  
  /* For getting pixmap we must have a window */
  gtk_widget_realize(window);
  
  play = gtk_toggle_button_new();  
  /* This calls our box creating function */
  box1 = xpm_box(window, PLAY_XPM);
  
  /* Pack and show all our widgets */
  gtk_widget_show(box1);
  gtk_container_add (GTK_CONTAINER (play), box1);
  
  gtk_signal_connect (GTK_OBJECT (play), "clicked",
		      GTK_SIGNAL_FUNC (play_status), (gpointer)PLAY_STRING);
  

  gtk_box_pack_start(GTK_BOX(hbox), play, FALSE, FALSE, 3);
  gtk_widget_show(play);
  
  pause = gtk_toggle_button_new();  
  box1 = xpm_box(window, PAUSE_XPM);

  /* Pack and show all our widgets */
  gtk_widget_show(box1);
  gtk_container_add (GTK_CONTAINER (pause), box1);

  gtk_signal_connect (GTK_OBJECT (pause), "clicked",
		      GTK_SIGNAL_FUNC (play_status), (gpointer)PAUSE_STRING);

  gtk_box_pack_start(GTK_BOX(hbox), pause, FALSE, FALSE, 3);
  gtk_widget_show(pause);

      
  gtk_widget_show(hbox);
  
  /* Show the main vbox and window */
  gtk_widget_show (vbox);
  gtk_widget_show (window);
  
  setPlayerStatus(PLAYING);

  /* THE LOOP */
  while (!done) {
    
    if (IS_PLAYING(player_status)) { /* Playing */

      /* Video Part */
      if (have_video) {
	
	// long sample, atime; 
	
	/* Restart clock */
	if (prev_frame == -1) {

	  (decoder->getFrame)(m_fp, &prev_frame);  
	  current_time = (long)(prev_frame * 1000.00/frate);
	  ref_start  = getTicks();

	  frames = 0;         /* Frame rate   */
	  fstart = getTicks();/* calculations */  
	}
	
	
	// (decoder->getSample)(m_fp, &sample);
	// atime = (long)((double)sample * 1000/spec.freq);

	
	/* Video Sync-ing
	 * Syncs with time - Not audio
	 * Assumption - Audio is going on well and sync-ed with time
	 */	
	current_frame = (long)(((double)frate * 
				(current_time + getTicks() - ref_start))
			       / 1000);
	
	/* Do video if it's not ahead of time */
	if (current_frame > prev_frame) {		
	  
	  /* Video Frame Draw Time */
	  start_time = getTicks();       
	  
	  /* Jump ahead if the video is behind time */
	  if (current_frame != prev_frame + 1) {
	    
	    /*** Review - Only if plugin allows it? 
	     *** Should plugin be given this choice ?
	     ***/
	    if ((decoder->isSeekable)(m_fp)) {
	      (decoder->setFrame)(m_fp, current_frame);
	    }
	  }
	  
	  /* Decode into the screen array and update the Drawing Area */
	  DRAW_VIDEO_FRAME;
	  
	  /* Do GTK events processing */
	  GTK_ITERATION_DO;
	  
	  /* Video Frame Draw (And GTK Event Iter) Time */
	  stop_time = getTicks();
	  
	  /* Note : diff is intentionally not updated if video is ahead */
	  fdecode +=  stop_time - start_time;
	  
	  /* Frame done */
	  prev_frame = current_frame;
	  frames++;
	  
	  if (getTicks() - fstart > 1000) {

#ifdef __DEBUG
	    printf("No. of frames per second : %3.2f\n", 
		   (frames * 1000.00)/(getTicks() - fstart));
#else
	    printf("\rNo. of frames per second : %3.2f", 
		   (frames * 1000.00)/(getTicks() - fstart));
	    fflush(stdout);
#endif
	    
	    frames = 0;
	    fstart = getTicks();
	  }
	}

	
	if (have_audio) {
	  
	  old_write_samples = spec.write_samples;
	  
	  /** Calculate audio decode sample size **/
	  
	  /* The equation here is :
	   * (Time taken to play the audio sample) >=
	   *      (Time taken to decode the audio sample) +
	   *      (Time spent in decoding and displaying the last frame)
	   */
	  
	  /* Video frame decode time is in fdecode*/
	  spec.write_samples = (int)(fdecode * spec.freq)
	    / (1000 - (test_decode * spec.freq/TEST_NUM_SAMPLES));
	  
	  /* Bring to the next nearest multiple of  STEP_AUDIO_SAMPLES */
	  spec.write_samples = ((spec.write_samples & 
				 (0xFFFF - STEP_AUDIO_SAMPLES)) +
				STEP_AUDIO_SAMPLES);
	  
	  if (spec.write_samples < MIN_AUDIO_SAMPLES)
	    spec.write_samples = MIN_AUDIO_SAMPLES;
	  
	  if (spec.write_samples > MAX_AUDIO_SAMPLES)
	    spec.write_samples = MAX_AUDIO_SAMPLES;
	  
#ifdef __DEBUG
	  adecode = (test_decode * spec.write_samples)/TEST_NUM_SAMPLES;
	  agap    = (spec.write_samples * 1000)/spec.freq;
	    
	  printf("\nSample size = %d\n", spec.write_samples);
	  printf("Time taken to decode 1 frame = %d\n", fdecode); 	  
	  printf("Time taken to decode 1 frag  = %d\n", adecode);
	  printf("Time taken to play   1 frag  = %d\n", agap);
#endif
	  
	  /* Decode size changed ? */
	  /* Tell audio device about change, so that it can adjust if reqd */
	  if (old_write_samples != spec.write_samples) {	    	    
	    setAudioWriteSamples(&spec);	    
	  }
	  
	  /* Reset video decode aggregate */
	  fdecode = 0;	
	}
      }

      /* Play audio if it's there */
      if (have_audio) {
	
	/* Pause status could have changed in the middle - Check again */
	if (IS_PLAYING(player_status)) {

	  /* Decode audio */
	  DECODE_AUDIO(m_fp, mixbuf, spec.write_samples);	  
	  
	  /* Play audio */ 
	  playAudio(audio_fd, mixbuf, spec.write_size);
	  
	  if (!have_video) {
	    /* Do GTK events processing if only audio is present */
	    GTK_ITERATION_DO;
	  }
	}	
      }

      if (!done) {
	/* Update slider position without generating "value_changed"
	 * "value_changed" should be generated only when the user drags the 
	 * slider, not when we update it ourselves
	 */
	adjustment = gtk_range_get_adjustment(GTK_RANGE(slider));
	
	pos = (float)getPosition();
	//printf("pos : %3.0f\n", pos);	       

	adjustment->value = pos;
	gtk_range_set_adjustment( GTK_RANGE(slider), adjustment);
	gtk_signal_emit_by_name (GTK_OBJECT (adjustment), "changed");
	
	/* Are we through? */
	if (have_video) {
	  if (pos >= total_frames) {
	    setPlayerStatus(PAUSED);
	  }
	}
	else {	  
	  if (pos >= total_samples) {
	    setPlayerStatus(PAUSED);
	  }
	}
      }
    }
    else { /* Everything's Paused */      

      if (!done) {
	/* Just do GTK event procesing */
	GTK_ITERATION_DO;      
      }      

      /* Sleep for sometime while paused */
      usleep(10000);
    }
  } /* while */

  printf("\n");
  
  /* Close audio device if reqd */
  if (have_audio) {
    free(mixbuf);
    closeAudio(audio_fd);
  }
  
  /* Close the file handle */
  (decoder->closeHandle)(m_fp);
  
  /* Free pixel buffers */
  free(pixbuf);
  free(pixel_rows);
  
  /* Unload plugin */
  unloadDecoderPlugin(plugin);

  return 0;
}

