/* libmpeg3_decoder.c 
 *
 * Implementation of the FamDecoder interface to provide libmpeg3 decoding 
 * functions. Handles only one stream of video and audio.
 *
 */
#include <stdio.h>
#include <stdlib.h>

#include "libmpeg3.h"
#include "famdecoderplugin.h"

/* Video decoding Color model  */
#define COLOR_MODEL  MPEG3_RGB888

static char decoderName[] = "libmpeg3";

/* Put extern "C" So that dlsym can find the entry points */
extern "C" {

  /* Plugin function prototypes */
  
  /** COMMON **/
  /* Return a name for the decoder plugin */
  const char *libmpeg3_decoderName(void); 

  /* Opens the file and returns a decoder specific handle */
  void *libmpeg3_openFile(char *path);
  
  /* Close the decoder specific handle */
  void libmpeg3_closeHandle(void *handle);
  
  /** VIDEO **/
  /* Tells whether a video stream is present */
  int libmpeg3_hasVideo(void *handle);
  
  /* Get video dimensions */
  void  libmpeg3_getVideoSize(void *handle, int *width, int *height);
  
  /* Get video frame rate */
  void  libmpeg3_getVideoFrameRate(void *handle, float *frame_rate);
  
  /* Get total no. of video frames */
  void  libmpeg3_getVideoFrames(void *handle, long *frames);
  
  /* Decode function - Color model should be RGB565 */
  void  libmpeg3_readVideoFrame(void *handle, 
				unsigned char **output_rows,
				int in_x,                   
				int in_y, 
				int in_w, 
				int in_h, 
				int out_w,          
				int out_h);  
  
  
  /** AUDIO **/
  /* Tells whether an audio stream is present */
  int   libmpeg3_hasAudio(void *handle);
  
  /* Get Sample rate */
  void  libmpeg3_getAudioSampleRate(void *handle, int *sample_rate);
  
  /* Get the number of audio channels - 1 or 2 */
  void  libmpeg3_getAudioChannels(void *handle, int *channels);
  
  /* Get total number of samples */
  void  libmpeg3_getAudioSamples(void *handle, long *num_samples);
  
  /* Decode function with 16-bit samples */ 
  void  libmpeg3_readAudioSamples(void *handle, short *buffer, 
				  long num_samples);
  
  /** SEEK **/
  /* Whether the decoder plugin can support frame and sample seeking */
  int libmpeg3_isSeekable(void *handle);
  
  /* Get the current audio position in samples */
  void libmpeg3_getSample(void *handle, long *sample);
  
  /* Set the current audio sample position in samples */
  void libmpeg3_setSample(void *handle, long sample);
  
  /* Get the current video position in frames */
  void libmpeg3_getFrame(void *handle, long *frame);
  
  /* Set the current video poisiton in frames */
  void libmpeg3_setFrame(void *handle, long frame);
    
  
  /** initDecoder **/
  /* All decoder plugins MUST have this function */
  FamDecoder *initDecoder(char *file_path)
  {
    FamDecoder *decoder;
    
    /* See whether we can handle the media file */
    if (mpeg3_check_sig(file_path)) {
      
      /* Initialize the decoder and send it across */
      decoder = (FamDecoder *)malloc(sizeof(FamDecoder));
      
      /* Initialize all the decoder vectors */
      decoder->decoderName        = libmpeg3_decoderName;
      decoder->openFile           = libmpeg3_openFile;
      decoder->closeHandle        = libmpeg3_closeHandle;
      
      decoder->hasVideo           = libmpeg3_hasVideo;
      decoder->getVideoSize       = libmpeg3_getVideoSize;
      decoder->getVideoFrameRate  = libmpeg3_getVideoFrameRate;
      decoder->getVideoFrames     = libmpeg3_getVideoFrames;
      decoder->readVideoFrame     = libmpeg3_readVideoFrame;
      
      decoder->hasAudio           = libmpeg3_hasAudio;
      decoder->getAudioSampleRate = libmpeg3_getAudioSampleRate;
      decoder->getAudioChannels   = libmpeg3_getAudioChannels;
      decoder->getAudioSamples    = libmpeg3_getAudioSamples;
      decoder->readAudioSamples   = libmpeg3_readAudioSamples;
      
      decoder->isSeekable         = libmpeg3_isSeekable;
      decoder->getSample          = libmpeg3_getSample;
      decoder->setSample          = libmpeg3_setSample;
      decoder->getFrame           = libmpeg3_getFrame;
      decoder->setFrame           = libmpeg3_setFrame;
      
#ifdef __DEBUG
      printf("%s loaded\n", decoderName);
#endif
      
      return decoder;
    }
    else {
      return NULL;
    }
  }

  /** destroyDecoder **/
  /* Optional Plugin Destructor */
  void destroyDecoder(FamDecoder *decoder)
  {
    free(decoder);
    
#ifdef __DEBUG
    printf("%s unloaded\n", decoderName);
#endif
  }
  
  
  /** COMMON **/
  /* Return a name for the decoder plugin */
  const char *libmpeg3_decoderName(void)
  {
    return (const char *)decoderName;
  }
  
  /* Opens the file and returns a decoder specific handle */
  void *libmpeg3_openFile(char *path)
  {
    mpeg3_t *m_fp;  
    m_fp = mpeg3_open(path);
    return (void *)m_fp;
  }
  
  /* Close the decoder specific handle */
  void libmpeg3_closeHandle(void *handle)
  {
    mpeg3_close((mpeg3_t *)handle);
  }
  
  /** VIDEO **/
  /* Tells whether a video stream is present */
  int libmpeg3_hasVideo(void *handle)
  {
    return mpeg3_has_video((mpeg3_t *)handle);
  }
  
  /* Get video dimensions */
  void  libmpeg3_getVideoSize(void *handle, int *width, int *height)
  {
    *width  = mpeg3_video_width((mpeg3_t *)handle, 0);
    *height = mpeg3_video_height((mpeg3_t *)handle, 0);  
  }
  
  /* Get video frame rate */
  void  libmpeg3_getVideoFrameRate(void *handle, float *frame_rate)
  {
    *frame_rate = mpeg3_frame_rate((mpeg3_t *)handle, 0);
  }
  
  /* Get total no. of video frames */
  void  libmpeg3_getVideoFrames(void *handle, long *frames)
  {
    *frames = mpeg3_video_frames((mpeg3_t *)handle, 0);
  }
  
  /* Decode function - Color model should be RGB565 */
  void  libmpeg3_readVideoFrame(void *handle, 
				unsigned char **output_rows,
				int in_x,                   
				int in_y, 
				int in_w, 
				int in_h, 
				int out_w,          
				int out_h)
  {
    mpeg3_read_frame((mpeg3_t *)handle, output_rows, 
		     in_x, in_y, in_w, in_h, 
		     out_w, out_h, 
		     COLOR_MODEL, 0);  
  }
  
  /** AUDIO **/
  /* Tells whether an audio stream is present */
  int   libmpeg3_hasAudio(void *handle)
  {
    return mpeg3_has_audio((mpeg3_t *)handle);
  }

  /* Get Sample rate */
  void  libmpeg3_getAudioSampleRate(void *handle, int *sample_rate)
  {
    *sample_rate = mpeg3_sample_rate((mpeg3_t *)handle, 0);
  }
  
  /* Get the number of audio channels - 1 or 2 */
  void  libmpeg3_getAudioChannels(void *handle, int *channels)
  {
    *channels = mpeg3_audio_channels((mpeg3_t *)handle, 0);
  }
  
  /* Get total number of samples */
  void  libmpeg3_getAudioSamples(void *handle, long *num_samples)
  {
    *num_samples = mpeg3_audio_samples((mpeg3_t *)handle, 0);
  }
  
  /* Decode function with 16-bit samples */ 
  void  libmpeg3_readAudioSamples(void *handle, short *buffer, 
				  long num_samples)
  {
    int channels = mpeg3_audio_channels((mpeg3_t *)handle, 0);
    
    if (channels == 1) { /* Mono */
      mpeg3_read_audio((mpeg3_t *)handle, NULL, buffer, 0, 0, num_samples, 0); 
    }
    else {               /* Stereo */
   
      mpeg3_read_audio((mpeg3_t *)handle, NULL, buffer, 1, 0, 
		       num_samples, 0); 
      mpeg3_reread_audio((mpeg3_t *)handle, NULL, buffer + 1, 1, 1,
			 num_samples, 0); 
    }    
  }
  
  /** SEEK **/
  /* Whether the decoder plugin can support frame and sample seeking */
  int libmpeg3_isSeekable(void *handle)
  {
    return 1; /* Always seekable, bad sometimes?? */
  }

  void libmpeg3_getSample(void *handle, long *sample)
  {
    *sample = mpeg3_get_sample((mpeg3_t *)handle, 0);
  }
  
  /* Set the current audio sample position in samples */
  void libmpeg3_setSample(void *handle, long sample)
  {
    mpeg3_set_sample((mpeg3_t *)handle, sample, 0);
  }
  
  /* Get the current video position in frames */
  void libmpeg3_getFrame(void *handle, long *frame)
  {
    *frame = mpeg3_get_frame((mpeg3_t *)handle, 0);
  }
  
  /* Set the current video poisiton in frames */
  void libmpeg3_setFrame(void *handle, long frame)
  {
    mpeg3_set_frame((mpeg3_t *)handle, frame, 0);
  }
  
};
