/* vorbis_decoder.c 
 *
 * Implementation of the FamDecoder interface to provide fixpvorbis decoding 
 * functions.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "famdecoderplugin.h"

#include "vorbis_codec.h"
#include "vorbisfile.h"

/* Bytes per audio sample */
#define SAMPLE_SIZE 2

static char decoderName[] = "fixpvorbis";

/* Bind few things that we have to maintain per stream */
typedef struct _vorbis_hanlde {
  
  OggVorbis_File *v_fp;
  vorbis_info    *vi;
  int            current_section; /* What are we supposed to do with this?? */

} VorbisHandle;


/* Plugin function prototypes */

/** COMMON **/
/* Return a name for the decoder plugin */
const char *vorbis_decoderName(void); 

/* Opens the file and returns a decoder specific handle */
void *vorbis_openFile(char *path);

/* Close the decoder specific handle */
void vorbis_closeHandle(void *handle);

/** VIDEO **/
/* Tells whether a video stream is present */
/* Always return FALSE for vorbis */
int vorbis_hasVideo(void *handle);

/** AUDIO **/
/* Tells whether an audio stream is present */
int   vorbis_hasAudio(void *handle);

/* Get Sample rate */
void  vorbis_getAudioSampleRate(void *handle, int *sample_rate);

/* Get the number of audio channels - 1 or 2 */
void  vorbis_getAudioChannels(void *handle, int *channels);

/* Get total number of samples */
void  vorbis_getAudioSamples(void *handle, long *num_samples);

/* Decode function with 16-bit samples */ 
void  vorbis_readAudioSamples(void *handle, short *buffer, 
				long num_samples);

/** SEEK **/
/* Whether the decoder plugin can support audio sample seeking */
int vorbis_isSeekable(void *handle);
     
/* Get the current audio position in samples */
void vorbis_getSample(void *handle, long *sample);

/* Set the current audio sample position in samples */
void vorbis_setSample(void *handle, long sample);

/** initDecoder **/
/* All decoder plugins MUST have this function */
FamDecoder *initDecoder(char *file_path)
{
  FILE *fp;
  OggVorbis_File vf;
  
  fp = fopen(file_path, "r");
  if (fp == NULL) 
    return NULL;

  /* See whether we can handle the media file */
  if (ov_open(fp, &vf, NULL, 0) < 0) {
    fclose(fp);
    return NULL; /* Not a Ogg Vorbis stream */
  }
  else {
    FamDecoder *decoder;  
    
    /* Initialize the decoder and send it across */
    decoder = (FamDecoder *)malloc(sizeof(FamDecoder));
    
    /* Initialize all the decoder vectors */
    decoder->decoderName        = vorbis_decoderName;
    decoder->openFile           = vorbis_openFile;
    decoder->closeHandle        = vorbis_closeHandle;
    
    decoder->hasVideo           = vorbis_hasVideo;
    
    decoder->hasAudio           = vorbis_hasAudio;
    decoder->getAudioSampleRate = vorbis_getAudioSampleRate;
    decoder->getAudioChannels   = vorbis_getAudioChannels;
    decoder->getAudioSamples    = vorbis_getAudioSamples;
    decoder->readAudioSamples   = vorbis_readAudioSamples;
    
    decoder->isSeekable         = vorbis_isSeekable;
    decoder->getSample          = vorbis_getSample;
    decoder->setSample          = vorbis_setSample;

    
#ifdef __DEBUG
    printf("%s loaded\n", decoderName);
#endif

    ov_clear(&vf);
    return decoder;
  }
}

/** destroyDecoder **/
/* Optional Plugin Destructor */
void destroyDecoder(FamDecoder *decoder)
{
  free(decoder);
  
#ifdef __DEBUG
  printf("%s unloaded\n", decoderName);
#endif
}
  
  
/** COMMON **/
/* Return a name for the decoder plugin */
const char *vorbis_decoderName(void)
{
  return (const char *)decoderName;
}

/* Opens the file and returns a decoder specific handle */
void *vorbis_openFile(char *path)
{
  FILE *fp;
  VorbisHandle *handle;
  char **ptr;
  
  /* Open Ogg vorbis file and get a hndle to it */
  fp = fopen(path, "r");
  if (fp == NULL) {
    return NULL;
  }
  
  handle = (VorbisHandle *)malloc(sizeof(VorbisHandle));
  if (handle == NULL){
    return NULL;
  }

  handle->v_fp = (OggVorbis_File *)malloc(sizeof(OggVorbis_File));  
  if (handle->v_fp == NULL){
    return NULL;
  }
  
  if (ov_open(fp, handle->v_fp, NULL, 0) < 0) {
    return NULL;
  }
 
  /* Get necessary information from the file */
  handle->vi = ov_info(handle->v_fp,-1);
  ptr = ov_comment(handle->v_fp,-1)->user_comments;

  return (void *)handle;
}

/* Close the decoder specific handle */
void vorbis_closeHandle(void *handle)
{
  ov_clear(((VorbisHandle *)handle)->v_fp);
  free(((VorbisHandle *)handle)->v_fp);
  free(handle);
}

/** VIDEO **/
/* Tells whether a video stream is present */
int vorbis_hasVideo(void *handle)
{
  return 0; /* No video streams for Ogg */
}
  
/** AUDIO **/
/* Tells whether an audio stream is present */
int   vorbis_hasAudio(void *handle)
{
  /* We already check for signature on the file */
  return 1;
}

/* Get Sample rate */
void  vorbis_getAudioSampleRate(void *handle, int *sample_rate)
{
  *sample_rate = ((VorbisHandle *)handle)->vi->rate;
}

/* Get the number of audio channels - 1 or 2 */
void  vorbis_getAudioChannels(void *handle, int *channels)
{
  *channels = ((VorbisHandle *)handle)->vi->channels;
}

/* Get total number of samples */
void  vorbis_getAudioSamples(void *handle, long *num_samples)
{
  *num_samples = ov_pcm_total(((VorbisHandle *)handle)->v_fp,-1);
}

/* Decode function with 16-bit samples */ 
void  vorbis_readAudioSamples(void *handle, short *buffer, long num_samples)
{
  VorbisHandle   *vh;
  OggVorbis_File *v_fp;
  char *cbuffer;
  int length, read, n;

  cbuffer = (char *)buffer;
  vh      = ((VorbisHandle *)handle);
  v_fp    = vh->v_fp;
  length  = num_samples * vh->vi->channels * SAMPLE_SIZE;
  
  /* ov_read reads only one vorbis packet at a time */
  read = 0;
  do {
    n = ov_read(v_fp, cbuffer + read, length - read, 0, 2, 1,
		&(vh->current_section));    
    
    if (n > 0) 
      read += n;
    else 
      break; /* EOF or some error :-P */
    
  } while (read < length);
}

/** SEEK **/
/* Whether the decoder plugin can support audio sample seeking */
int vorbis_isSeekable(void *handle)
{
  return ov_seekable(((VorbisHandle *)handle)->v_fp);
}

/* Get the current audio position in samples */
void vorbis_getSample(void *handle, long *sample)
{
  *sample = (long)ov_pcm_tell(((VorbisHandle *)handle)->v_fp);
}

/* Set the current audio sample position in samples */
void vorbis_setSample(void *handle, long sample)
{
  ov_pcm_seek(((VorbisHandle *)handle)->v_fp, sample);
}
